Unit LFBProcs;

Interface

Uses
    Glide2x;

Const
   // Fehler-Konstanten
   LFB_Error: Integer = 0;
   LFB_NoError = 0;
   LFB_OutOfMemError = 1;
   LFB_IOError = 2;
   LFB_WrongFormatError = 3;
   LFB_AnyError = 4;

Type
    LFB_Sprite = Record
       Width, Height: Longint;
       Adr: Pointer;
    End;

Function LFB_Lock: Pointer;
Procedure LFB_Unlock;

Procedure LFB_PutPixel(x, y: Integer; Color: Word);
Procedure LFB_PutPixelSprite(x, y: Integer; Color: Word; Const Spr: LFB_Sprite);
Function LFB_GetPixelSprite(x, y: Integer; Const Spr: LFB_Sprite): Word;
Function LFB_Encode565(r, g, b: Byte): Word;
Procedure LFB_Decode565(Color: Word; Var r, g, b: Byte); Pascal;

Procedure LFB_LoadPCXFile(FileName: String; Var DestSprite: LFB_Sprite);
Procedure LFB_PutSprite(x, y: Integer; SourceSprite: LFB_Sprite);
Procedure LFB_PutSpriteFast(x, y: Integer; SourceSprite: LFB_Sprite);
Procedure LFB_GetSpriteFromSprite(Source: LFB_Sprite; Var Dest: LFB_Sprite; x, y, Width, Height: Longint);
Procedure LFB_FreeSprite(Var Sprite: LFB_Sprite);

Procedure LFB_SetClippingRect(x1, y1, x2, y2: Integer);
Procedure LFB_GetClippingRect(Var x1, y1, x2, y2: Integer);

Var
   LFB_Pointer: Pointer;
   LFB_StrideInBytes: Integer;
   cr_x1, cr_y1, cr_x2, cr_y2: Integer;

Implementation

Type
   PCXHeader = Record
        Manufacturer: byte;     { Always 10 for PCX file }

        Version: byte;          { 2 - old PCX - no palette (not used anymore),
                                  3 - no palette,
                                  4 - Microsoft Windows - no palette (only in
                                      old files, new Windows version uses 3),
                                  5 - with palette }

        Encoding: byte;         { 1 is PCX, it is possible that we may add
                                  additional encoding methods in the future }

        Bits_per_pixel: byte;   { Number of bits to represent a pixel
                                  (per plane) - 1, 2, 4, or 8 }

        Xmin: Word;          { Image window dimensions (inclusive) }
        Ymin: Word;          { Xmin, Ymin are usually zero (not always) }
        Xmax: Word;
        Ymax: Word;

        Hdpi: Word;          { Resolution of image (dots per inch) }
        Vdpi: Word;          { Set to scanner resolution - 300 is default }

        ColorMap: Array [0..15, 0..2] Of byte;
                                { RGB palette data (16 colors or less)
                                  256 color palette is appended to end of file }

        Reserved: byte;         { (used to contain video mode)
                                  now it is ignored - just set to zero }

        Nplanes: byte;          { Number of planes }

        Bytes_per_line_per_plane: Word;   { Number of bytes to allocate
                                               for a scanline plane.
                                               MUST be an an EVEN number!
                                               Do NOT calculate from Xmax-Xmin! }

        PaletteInfo: Word;   { 1 = black & white or color image,
                                  2 = grayscale image - ignored in PB4, PB4+
                                  palette must also be set to shades of gray! }

        HscreenSize: Word;   { added for PC Paintbrush IV Plus ver 1.0,  }
        VscreenSize: Word;   { PC Paintbrush IV ver 1.02 (and later)     }
                                { I know it is tempting to use these fields
                                  to determine what video mode should be used
                                  to display the image - but it is NOT
                                  recommended since the fields will probably
                                  just contain garbage. It is better to have
                                  the user install for the graphics mode he
                                  wants to use... }

        Filler: Array [74..127] Of byte;     { Just set to zeros }
   End;

Var
   LFBInfo: TgrLFBInfo;

Function LFB_Lock;
Begin
     LFBInfo.Size:=SizeOf(TgrLFBInfo);
     If grLFBLock(GR_LFB_WRITE_ONLY Or GR_LFB_NOIDLE, GR_BUFFER_BACKBUFFER, GR_LFBWRITEMODE_565, GR_ORIGIN_UPPER_LEFT, FXFALSE, LFBInfo) Then Begin
        LFB_Pointer:=LFBInfo.lfbPtr;
        LFB_StrideInBytes:=LFBInfo.strideInBytes;
     End;
     Result:=LFB_Pointer;
End;

Procedure LFB_Unlock;
Begin
     grLFBUnlock(GR_LFB_WRITE_ONLY Or GR_LFB_NOIDLE, GR_BUFFER_BACKBUFFER);
     LFB_Pointer:=Nil;
     LFB_StrideInBytes:=0;
End;

Procedure LFB_PutPixel;
Begin
     Asm
        pushad
        xor    edx, edx
        mov    eax, LFB_StrideInBytes
        mul    y
        mov    ebx, x
        shl    ebx, 1
        add    eax, ebx
        add    eax, LFB_Pointer
        mov    edi, eax
        mov    ax, Color
        mov    [edi], ax
        popad
     End;
End;

Function LFB_Encode565;
Begin
     Asm
        pushad
        xor    ax,ax
        mov    al,b
        xor    bx,bx
        mov    bl,g
        shl    bx,5
        or     ax,bx
        xor    bx,bx
        mov    bl,r
        shl    bx,11
        or     ax,bx
        mov    Result,ax
        popad
     End;
End;

Procedure LFB_LoadPCXFile;
Const
   Compress = $C0;
   DeCompress = $3F;
Var
   Buffer: Array[0..65535] Of Byte;
   Palette: Array[0..767] Of Byte;
   Colors: Array[0..255] Of Word;
   Header: PCXHeader;
   f: File;
   _Read: Integer;
   Width, Height: Integer;
   Size: Longint;
   i, OldPos: Longint;
   BufSize: Longint;
   Data, Counter: Byte;
   Offs: Longint;
   SourceOffs: Longint;
   ColAdr: Longint;
   ScanLineCount: Longint;

Function ReadNextByte: Byte;
Begin
     {$I-}
     If BufSize <= 0 Then Begin
        BlockRead(f,Buffer,SizeOf(Buffer),_Read);
        BufSize:=_Read;
        If (BufSize = 0) Or (IOResult <> 0) Then Begin
           LFB_Error:=LFB_IOError;
           Result:=0;
           Exit;
        End Else Begin
           SourceOffs:=0;
           Result:=ReadNextByte;
        End;
     End Else Begin
        Result:=Buffer[SourceOffs];
        Dec(BufSize);
        Inc(SourceOffs);
     End;
     {$I+}
End;

Begin
     {$I-}

     // Datei ffnen, abbrechen falls Fehler
     IOResult;
     AssignFile(f, FileName);
     Reset(f, 1);
     If IOResult <> 0 Then Begin
        LFB_Error:=LFB_IOError;
        Exit;
     End;

     // Header einlesen, abberechen falls Fehler
     BlockRead(f, Header, SizeOf(Header), _Read);
     If (_Read <> SizeOf(Header)) Or (IOResult <> 0) Then Begin
        CloseFile(f);
        IOResult;
        LFB_Error:=LFB_IOError;
        Exit;
     End;

     // Speicher berechnen und allokieren, abbrechen falls Fehler
     With Header Do Begin
          Width:=Abs(XMax-XMin)+1;
          Height:=Abs(YMax-YMin)+1;
          If (Bits_Per_Pixel <> 8) Or (NPLanes <> 1) Or (Width = 0)
             Or (Height = 0) Then Begin
             CloseFile(f);
             IOResult;
             LFB_Error:=LFB_WrongFormatError;
             Exit;
          End;
     End;
     Size:=Longint(Width)*Longint(Height);
     GetMem(DestSprite.Adr, Size*2);
     DestSprite.Width:=Width;
     DestSprite.Height:=Height;

     // Palette einlesen abbrechen, falls Fehler
     OldPos:=FilePos(f);
     Seek(f,FileSize(f)-SizeOf(Palette));
     BlockRead(f,Palette,SizeOf(Palette),_Read);
     If (IOResult <> 0) Or (_Read <> SizeOf(Palette)) Then Begin
        FreeMem(DestSprite.Adr,Size*2);
        CloseFile(f);
        IOResult;
        LFB_Error:=LFB_IOError;
        Exit;
     End;
     Seek(f,OldPos);

     // Palette in 16-Bit Farbwerte umrechnen
     For i:=0 To 255 Do Begin
         Colors[i]:=LFB_Encode565(Palette[i*3] Shr 3,
                                  Palette[i*3+1] Shr 2,
                                  Palette[i*3+2] Shr 3);
     End;
     // Variablen initialisieren
     i:=Size;
     BufSize:=0;
     Offs:=Longint(DestSprite.Adr);
     SourceOffs:=0;
     ColAdr:=Longint(@Colors);
     ScanLineCount:=DestSprite.Width;

     // Jetzt alle Bytes durchgehen und prfen, ob die beiden
     // obersten Bit gesetzt sind. Ist das der Fall, ist das Byte
     // ein Zhl-Byte
     Repeat
           Data:=ReadNextByte;
           If LFB_Error = LFB_IOError Then Begin
              FreeMem(DestSprite.Adr,Size*2);
              CloseFile(f);
              IOResult;
              Exit;
           End;
           If (Data And Compress) = Compress Then Begin
              Counter:=(Data And Decompress);
              Data:=ReadNextByte;
              If LFB_Error = LFB_IOError Then Begin
                 FreeMem(DestSprite.Adr,Size*2);
                 CloseFile(f);
                 IOResult;
                 Exit;
              End;
           End Else Counter:=1;
           Asm
              pushad
              // Farbwert aus Tabelle lesen
              mov     edi, ColAdr
              mov     eax, 0
              mov     al, Data
              shl     eax, 1
              add     edi, eax
              mov     ax, [edi]
              // Zieladresse und Zhler laden
              mov     edi, Offs
              mov     cl, Counter
              or      cl, cl
              je      @End
           @WriteLoop:
              // Eine Scanline mu ja immer "even" sein. Grummel, grummel!
              cmp     ScanLineCount, 0
              jne     @NoScanLineEnd
              mov     ebx, Width
              mov     ScanLineCount, ebx
              cmp     bx, Word Ptr Header.Bytes_Per_Line_Per_Plane
              jne     @OverridePixel
              // Pixel schreiben und Zieladdresse um 1 Word erhhen
           @NoScanLineEnd:
              mov     [edi], ax
              add     edi, 2
              add     Offs, 2
              dec     ScanLineCount
           @OverridePixel:
              dec     i
              dec     cl
              // Solange, bis CL = 0
              jne     @WriteLoop
           @End:
              popad
           End;
     Until i <= 0;

     // Ende - Datei schlieen
     CloseFile(f);
     IOResult;
     {$I+}

End;

Procedure LFB_PutSpriteFast;
Var
   x1, y1, x2, y2: Integer;
   Width, Height: Integer;
   StartX, StartY: Integer;
Begin
     // Ist das Sprite berhaupt im Bild? Wenn nicht -> Ende
     If (x > cr_x2) Or (y > cr_y2) Or (x+SourceSprite.Width < cr_x1) Or (y+SourceSprite.Height < cr_y1) Then Exit;

     // Ersteinmal annehmen, da alles sichtbar ist
     StartX:=x;
     StartY:=y;
     x1:=0;
     y1:=0;
     x2:=SourceSprite.Width-1;
     y2:=SourceSprite.Height-1;

     // Ggf. Rnder beschneiden
     If x < cr_x1 Then Begin
        Inc(x1, cr_x1 - x);
        StartX:=cr_x1;
     End;
     If y < cr_y1 Then Begin
        Inc(y1, cr_y1 - y);
        StartY:=cr_y1;
     End;
     If x+SourceSprite.Width > cr_x2 Then Dec(x2, x+SourceSprite.Width - cr_x2 - 1);
     If y+SourceSprite.Height > cr_y2 Then Dec(y2, y+SourceSprite.Height - cr_y2 - 1);

     // Breite und Hhe berechnen
     Width:=x2 - x1 + 1;
     Height:=y2 - y1 + 1;

     // Zeichnen
     grLFBWriteRegion(GR_BUFFER_BACKBUFFER, StartX, StartY, GR_LFB_SRC_FMT_565,
                      Width, Height, SourceSprite.Width*2,
                      Pointer(Integer(SourceSprite.Adr) + (y1*SourceSprite.Width+x1)*2));
End;

Procedure LFB_PutSprite;
Var
   x1, y1, x2, y2: Integer;
   Width, Height: Integer;
   StartX, StartY: Integer;
   Skip, SkipDest: Integer;
Begin
     // Ist das Sprite berhaupt im Bild? Wenn nicht -> Ende
     If (x > cr_x2) Or (y > cr_y2) Or (x+SourceSprite.Width < cr_x1) Or (y+SourceSprite.Height < cr_y1) Then Exit;

     // Ersteinmal annehmen, da alles sichtbar ist
     StartX:=x;
     StartY:=y;
     x1:=0;
     y1:=0;
     x2:=SourceSprite.Width-1;
     y2:=SourceSprite.Height-1;

     // Ggf. Rnder beschneiden
     If x < cr_x1 Then Begin
        Inc(x1, cr_x1 - x);
        StartX:=cr_x1;
     End;
     If y < cr_y1 Then Begin
        Inc(y1, cr_y1 - y);
        StartY:=cr_y1;
     End;
     If x+SourceSprite.Width > cr_x2 Then Dec(x2, x+SourceSprite.Width - cr_x2 - 1);
     If y+SourceSprite.Height > cr_y2 Then Dec(y2, y+SourceSprite.Height - cr_y2 - 1);

     // Breite und Hhe berechnen
     Width:=x2 - x1 + 1;
     Height:=y2 - y1 + 1;
     Skip:=(SourceSprite.Width-Width)*2;
     SkipDest:=LFB_StrideInBytes-(SourceSprite.Width*2-Skip);

     Asm
        pushad

        // Offset im Sprite berechnen
        xor     edx, edx
        mov     eax, DWord Ptr SourceSprite.Width
        mul     y1
        add     eax, x1
        shl     eax, 1
        add     eax, DWord Ptr SourceSprite.Adr
        mov     esi, eax

        // Offset im Buffer berechnen
        xor     edx, edx
        mov     eax, LFB_StrideInBytes
        mul     StartY
        mov     ebx, StartX
        shl     ebx, 1
        add     eax, ebx
        add     eax, LFB_Pointer
        mov     edi, eax

        // Zeichnen
        mov     edx, Height
     @YLoop:
        mov     ecx, Width
     @XLoop:
        mov     ax, [esi]
        or      ax, ax
        je      @NextPixel
        mov     [edi], ax
     @NextPixel:
        add     esi, 2
        add     edi, 2
        dec     ecx
        jne     @XLoop
        add     esi, Skip
        add     edi, SkipDest
        dec     edx
        jne     @YLoop

        popad
     End;
End;

Procedure Swap(Var i1, i2: Integer);
Var
   Dummy: Integer;
Begin
     Dummy:=i1;
     i1:=i2;
     i2:=Dummy;
End;

Procedure LFB_SetClippingRect;
Begin
     If x1 > x2 Then Swap(x1, x2);
     If y1 > y2 Then Swap(y1, y2);
     cr_x1:=x1;
     cr_y1:=y1;
     cr_x2:=x2;
     cr_y2:=y2;
End;

Procedure LFB_GetClippingRect;
Begin
     x1:=cr_x1;
     y1:=cr_y1;
     x2:=cr_x2;
     y2:=cr_y2;
End;

Procedure LFB_GetSpriteFromSprite;
Var
   Size: Longint;
   StartAdr: Longint;
   Skip: Longint;
   Sack: Longint;
Begin
     Size:=(Width*Height) * 2;
     GetMem(Dest.Adr, Size);
     StartAdr:=(y*Source.Width+x) * 2 + Longint(Source.Adr);
     Skip:=(Source.Width-Width) * 2;
     Sack:=Longint(Dest.Adr);
     Dest.Width:=Width;
     Dest.Height:=Height;
     Asm
        pushad
        mov    esi, StartAdr
        mov    edi, Sack
        mov    edx, Height
     @YLoop:
        mov    ecx, Width
     @XLoop:
        mov    ax, Word Ptr [esi]
        mov    [edi], ax
        add    edi, 2
        add    esi, 2
        dec    ecx
        jne    @XLoop
        add    esi, Skip
        dec    edx
        jne    @YLoop
        popad
     End;
End;

Procedure LFB_FreeSprite;
Begin
     If Sprite.Width > 0 Then Begin
        FreeMem(Sprite.Adr);
        Sprite.Adr:=Nil;
        Sprite.Width:=0;
        Sprite.Height:=0;
     End;
End;

Procedure LFB_PutPixelSprite;
Var
   Width: Integer;
   Adr: Pointer;
Begin
     Width:=Spr.Width;
     Adr:=Spr.Adr;
     Asm
        pushad
        xor    edx, edx
        mov    eax, Width
        shl    eax, 1
        mul    y
        mov    ebx, x
        shl    ebx, 1
        add    eax, ebx
        add    eax, Adr
        mov    edi, eax
        mov    ax, Color
        mov    [edi], ax
        popad
     End;
End;

Function LFB_GetPixelSprite;
Var
   Width: Integer;
   Adr: Pointer;
   Res: Word;
Begin
     Width:=Spr.Width;
     Adr:=Spr.Adr;
     Asm
        pushad
        xor    edx, edx
        mov    eax, Width
        shl    eax, 1
        mul    y
        mov    ebx, x
        shl    ebx, 1
        add    eax, ebx
        add    eax, Adr
        mov    esi, eax
        mov    ax, [esi]
        mov    Res, ax
        popad
     End;
     Result:=Res;
End;

Procedure LFB_Decode565; Pascal;
Begin
     Asm
        pushad
        mov    dx,Color
        mov    ax,dx
        and    al,00011111b
        mov    ebx,b
        mov    [ebx],al
        and    ah,11111000b
        shr    ah,3
        mov    ebx,r
        mov    [ebx],ah
        and    dx,0000011111100000b
        shr    dx,5
        mov    ebx,g
        mov    [ebx],dl
        popad
     End;
End;

Initialization
   LFB_Pointer:=Nil;
   LFB_StrideInBytes:=0;
   LFB_SetClippingRect(0, 0, 0, 0);
End.
